BossList = {}
local AceGUI = LibStub("AceGUI-3.0")

local dateFormat = "%m/%d/%y %H:%M:%S"
local difficultyNames = {"Normal", "Heroic", "Flex"}
difficultyNames[0] = ""
local desireOptions = {"BiS", "Upgrade", "Off-spec", "Not Needed"}
local desireColors = {{r=0, g=1, b=0}, {r=1, g=1, b=0}, {r=1, g=0.5, b=0}, {r=1, g=0, b=0}}
local tab
local scrollFrame
local instanceDropdown
local encounterDropdown
local curInstance, curDifficulty, curEncounter


--local BossListFrame = CreateFrame("Frame", "BossListFrame")
local BossListFrame = {}
BossList.frame = BossListFrame

local function pairsByKeys (t, f)
	local a = {}
		for n in pairs(t) do table.insert(a, n) end
		table.sort(a, f)
		local i = 0      -- iterator variable
		local iter = function ()   -- iterator function
			i = i + 1
			if a[i] == nil then return nil
			else return a[i], t[a[i]]
			end
		end
	return iter
end

local function TimeDifference(lastUpdate)
  local dif = time() - lastUpdate
  if dif < 0 then
    dif = 0
  end
  
  local dateDif = date("*t", dif)
  
  if dateDif.month > 1 then
    return string.format("%d month(s) ago", dateDif.month - 1)
  end
  
  if dateDif.day > 1 then
    return string.format("%d day(s) ago", dateDif.day - 1)
  end
  
  if dateDif.hour > 1 then
    return string.format("%d hour(s) ago", dateDif.hour - 1)
  end
  
  if dateDif.min > 1 then
    return string.format("%d minute(s) ago", dateDif.min - 1)
  end
  
  return string.format("%d second(s) ago", dateDif.sec)
end

local function DrawItems(container, selectedInstance, selectedDifficulty, selectedEncounter)
  local DB = BossList.DB
  container:ReleaseChildren()
  local children = {}
  local characterTable = {}

  for instance, difficulties in pairs(BossList_entries) do            -- Instances
    if selectedInstance == nil or instance == selectedInstance then   -- Filter on instance
      EJ_SelectInstance(instance) -- Required since 5.4 to make EJ_GetEncounterInfoByIndex work correctly
      for difficulty, encounters in pairs(difficulties) do            -- Difficulties
         if (selectedDifficulty == nil or difficulty == selectedDifficulty) and 
            (selectedEncounter == nil or encounters[selectedEncounter] ~= nil) then 
          -- Filter on difficulty, and check whether or not the selected encounter is in this instance
          
          local isHeaderPrinted = false
          local bossIndex = 1;  -- Iterate over the bosses in the order they appear in the instance
          local encounterName, _, encounter, _, _ = EJ_GetEncounterInfoByIndex(bossIndex, instance)
          
          while encounter ~= nil do           -- Encounters
            local isEncounterPrinted = false
            local items = encounters[encounter]
            if items ~= nil then
              for item, desires in pairs(items) do
                local isItemPrinted = false
                local itemLabel
                if desires ~= nil then 
                  for desire, desireName in ipairs(desireOptions) do  -- Desires
                  
                    local characters = desires[desire]
                    -- Check if it exists and filter
                    if characters ~= nil and (selectedEncounter == nil or encounter == selectedEncounter) then 
                      wipe(characterTable)
                      
                      for character, lastUpdate in pairsByKeys(characters) do -- Characters
                        if not(DB.global.hideNotNeeded) or desire < 4 then
                          if not isHeaderPrinted then
                            local instanceName = EJ_GetInstanceInfo(instance)
                            local difficultyName = difficultyNames[difficulty]
                            
                            local instanceHeading = AceGUI:Create("Heading")
                            instanceHeading:SetFullWidth(true)
                            instanceHeading:SetText(instanceName .. " - " .. difficultyName)
                            tinsert(children, instanceHeading)
                            
                            isHeaderPrinted = true
                          end
                          if not isEncounterPrinted then
                            local encounterLabel = AceGUI:Create("Label")        
                            encounterLabel:SetColor(1,0.82,0)
                            encounterLabel:SetFullWidth(true)
                            encounterLabel.label:SetJustifyH("LEFT")
                            encounterLabel:SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE" )
                            encounterLabel:SetText(encounterName)
                            tinsert(children, encounterLabel)
                            
                            isEncounterPrinted = true
                          end

                          tinsert(characterTable, character)
                          
                        end
                      end
                      
                      if not(isItemPrinted) and next(characterTable) then
                        local _, itemLink = GetItemInfo(item)
                        local itemTexture = GetItemIcon(item)
                        local query = false
                        if itemLink == nil then
                          itemLink = "|cFFFF0000Retrieving Item Information|r"
                          query = true
                        end
                        
                        itemLabel = AceGUI:Create("InteractiveLabel")
                        itemLabel:SetRelativeWidth(1)
                        itemLabel.label:SetJustifyH("LEFT")
                        itemLabel:SetText(itemLink)
                        itemLabel:SetImage(itemTexture)
                        
                        itemLabel:SetCallback("OnEnter", function(self) 
                                                GameTooltip:SetOwner(self.frame, "ANCHOR_CURSOR")
                                                GameTooltip:SetItemByID(item)
                                                
                                                GameTooltip:Show()
                                              end)
                        itemLabel:SetCallback("OnLeave", function(self) 
                                                GameTooltip:Hide()
                                              end)
                        
                        tinsert(children, itemLabel)
                        if query then
                          BossList.QueryItem(item, itemLabel.label)
                        end
                        
                        isItemPrinted = true
                      end
                      
                      local color = desireColors[desire]
                      
                      local characterLabel = AceGUI:Create("Label")
                      characterLabel:SetRelativeWidth(0.33)
                      characterLabel:SetText(table.concat(characterTable, "\n"))
                      characterLabel:SetColor(color.r, color.g, color.b)
                      characterLabel.label:SetJustifyH("LEFT")
                      tinsert(children, characterLabel)
                      
                    end
                  end
                  if isItemPrinted then
                    local heading = AceGUI:Create("Heading")
                    heading:SetFullWidth(true)
                    --container:AddChild(heading)
                    tinsert(children, heading)
                  end
                end
              end
            end
            bossIndex = bossIndex + 1;
            encounterName, _, encounter, _, _ = EJ_GetEncounterInfoByIndex(bossIndex, instance)
          end
        end
      end
    end
  end
  container:AddChildren(unpack(children))
  
  -- If an instance was selected in the dungeon journal, set it back
  if EncounterJournal ~= nil and EncounterJournal.instanceID ~= nil then
    EJ_SelectInstance(EncounterJournal.instanceID)
  end
end

local function DrawEntries(container, selectedInstance, selectedDifficulty, selectedEncounter)
  local DB = BossList.DB
  container:ReleaseChildren()
  local children = {}
  local characterTable = {}
  local desireTable = {}
  local lastUpdateTable = {}
  local entriesTable = {}
  for instance, difficulties in pairs(BossList_entries) do            -- Instances
    if selectedInstance == nil or instance == selectedInstance then   -- Filter on instance
      EJ_SelectInstance(instance) -- Required since 5.4 to make EJ_GetEncounterInfoByIndex work correctly
      for difficulty, encounters in pairs(difficulties) do            -- Difficulties
         if (selectedDifficulty == nil or difficulty == selectedDifficulty) and 
            (selectedEncounter == nil or encounters[selectedEncounter] ~= nil) then 
          -- Filter on difficulty, and check whether or not the selected encounter is in this instance
          
          local isHeaderPrinted = false
          local bossIndex = 1;  -- Iterate over the bosses in the order they appear in the instance
          local encounterName, _, encounter, _, _ = EJ_GetEncounterInfoByIndex(bossIndex, instance)
          
          while encounter ~= nil do           -- Encounters
            local isEncounterPrinted = false
            local items = encounters[encounter]
            if items ~= nil and (selectedEncounter == nil or encounter == selectedEncounter) then
              wipe(entriesTable)
              for desire, desireName in ipairs(desireOptions) do          -- Desires
                wipe(characterTable)
                wipe(desireTable)
                wipe(lastUpdateTable)
                for item, desires in pairs(items) do                      -- Items
                  if desires ~= nil then 
                    
                    local characters = desires[desire]
                    -- Check if it exists
                    if characters ~= nil then 
                      
                      for character, lastUpdate in pairs(characters) do -- Characters
                        
                        if not(DB.global.hideNotNeeded) or desire < 4 then
                          if not isHeaderPrinted then
                            local instanceName = EJ_GetInstanceInfo(instance)
                            local difficultyName = difficultyNames[difficulty]
                            
                            local instanceHeading = AceGUI:Create("Heading")
                            instanceHeading:SetFullWidth(true)
                            instanceHeading:SetText(instanceName .. " - " .. difficultyName)
                            tinsert(children, instanceHeading)
                            
                            isHeaderPrinted = true
                          end
                          if not isEncounterPrinted then
                            local encounterLabel = AceGUI:Create("Label")        
                            encounterLabel:SetColor(1,0.82,0)
                            encounterLabel:SetFullWidth(true)
                            encounterLabel.label:SetJustifyH("LEFT")
                            encounterLabel:SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE" )
                            encounterLabel:SetText(encounterName)
                            tinsert(children, encounterLabel)
                            
                            isEncounterPrinted = true
                          end
                          
                          if entriesTable[character] == nil then
                            entriesTable[character] = {desireName, TimeDifference(lastUpdate)}
                          end
                        end
                      end
                    end
                  end
                end
                
                if next(entriesTable) ~= nil then -- Check if any characters were found for this encounter
                  for character, details in pairsByKeys(entriesTable) do
                    if details[1] == desireName then
                      tinsert(characterTable, character)
                      tinsert(desireTable, details[1])
                      tinsert(lastUpdateTable, details[2])
                    end
                  end
                  
                  if next(characterTable) ~= nil then -- Check if any characters were found for this desire
                    local color = desireColors[desire]
                    local characterLabel = AceGUI:Create("Label")
                    characterLabel:SetRelativeWidth(0.4)
                    characterLabel:SetText(table.concat(characterTable, "\n"))
                    characterLabel:SetColor(color.r, color.g, color.b)
                    characterLabel.label:SetJustifyH("LEFT")
                    tinsert(children, characterLabel)
                    
                    local desireLabel = AceGUI:Create("Label")
                    desireLabel:SetRelativeWidth(0.29)
                    desireLabel:SetText(table.concat(desireTable, "\n"))
                    desireLabel:SetColor(color.r, color.g, color.b)
                    desireLabel:SetPoint("RIGHT", characterLabel.frame, 0,0)
                    --desireLabel:SetImage("Interface/BUTTONS/UI-GroupLoot-Coin-Up")
                    desireLabel:SetImageSize(30,30)
                    desireLabel.label:SetJustifyH("CENTER")
                    tinsert(children, desireLabel)
                    
                    local lastUpdateLabel = AceGUI:Create("Label")
                    lastUpdateLabel:SetRelativeWidth(0.3)
                    lastUpdateLabel:SetText(table.concat(lastUpdateTable, "\n"))
                    lastUpdateLabel:SetPoint("RIGHT", desireLabel.frame, 0,0)
                    lastUpdateLabel.label:SetJustifyH("RIGHT")
                    tinsert(children, lastUpdateLabel)
                  end
                end
              end
              if isEncounterPrinted then
                local heading = AceGUI:Create("Heading")
                heading:SetFullWidth(true)
                --container:AddChild(heading)
                tinsert(children, heading)
              end
            end
            
            bossIndex = bossIndex + 1;
            encounterName, _, encounter, _, _ = EJ_GetEncounterInfoByIndex(bossIndex, instance)
          end
        end
      end
    end
  end
  container:AddChildren(unpack(children))
  
  -- If an instance was selected in the dungeon journal, set it back
  if EncounterJournal ~= nil and EncounterJournal.instanceID ~= nil then
    EJ_SelectInstance(EncounterJournal.instanceID)
  end
end

local function UpdateInstanceList()
  local instances = {}
  for instance, _ in pairs(BossList_entries) do
    instances[instance] = EJ_GetInstanceInfo(instance)
  end
  instances[0] = ""

  instanceDropdown:SetList(instances)
end

local function UpdateEncounterList(instance)
  local encounterList = {}
  local orderList = {}
  encounterList[0] = ""
  tinsert(orderList, 0)
  if instance == 0 or instance == nil then -- Nothing selected
    curInstance = nil
    
    for instance, _ in pairs(BossList_entries) do -- Instances
      EJ_SelectInstance(instance)
      local bossIndex = 1;
      local encounterName, _, encounter, _, _ = EJ_GetEncounterInfoByIndex(bossIndex, instance)
      while encounter ~= nil do
        encounterList[encounter] = encounterName
        tinsert(orderList, encounter)
      
        bossIndex = bossIndex + 1;
        encounterName, _, encounter, _, _ = EJ_GetEncounterInfoByIndex(bossIndex, instance)
      end
    end

  elseif BossList_entries[instance] ~= nil then
    curInstance = instance
    EJ_SelectInstance(instance)
    local bossIndex = 1;
    local encounterName, _, encounter, _, _ = EJ_GetEncounterInfoByIndex(bossIndex, instance)
    while encounter ~= nil do
      encounterList[encounter] = encounterName
      tinsert(orderList, encounter)
    
      bossIndex = bossIndex + 1;
      encounterName, _, encounter, _, _ = EJ_GetEncounterInfoByIndex(bossIndex, instance)
    end
  end

  encounterDropdown:SetList(encounterList, orderList)
  encounterDropdown:SetValue(curEncounter)
  
  -- If an instance was selected in the dungeon journal, set it back
  if EncounterJournal ~= nil and EncounterJournal.instanceID ~= nil then
    EJ_SelectInstance(EncounterJournal.instanceID)
  end
end


local function FilterInstance(self, event, instance)
  
  curEncounter = nil
  
  UpdateEncounterList(instance)
  if tab.localstatus.selected == "tab1" then
    DrawEntries(scrollFrame, curInstance, curDifficulty, curEncounter)
  elseif tab.localstatus.selected == "tab2" then
    DrawItems(scrollFrame, curInstance, curDifficulty, curEncounter)
  end

end

local function FilterDifficulty(self, event, difficulty)
  if difficulty == 0 then
    curDifficulty = nil
  else
    curDifficulty = difficulty
  end
  if tab.localstatus.selected == "tab1" then
    DrawEntries(scrollFrame, curInstance, curDifficulty, curEncounter)
  elseif tab.localstatus.selected == "tab2" then
    DrawItems(scrollFrame, curInstance, curDifficulty, curEncounter)
  end
end

local function FilterEncounter(self, event, encounter)
  if encounter == 0 then
    curEncounter = nil
  else
    curEncounter = encounter
  end
  if tab.localstatus.selected == "tab1" then
    DrawEntries(scrollFrame, curInstance, curDifficulty, curEncounter)
  elseif tab.localstatus.selected == "tab2" then
    DrawItems(scrollFrame, curInstance, curDifficulty, curEncounter)
  end
end

local function DrawBossesTab(container)
  curDifficulty = nil
  curEncounter = nil
  curInstance = nil
  local difficultyDropdown = AceGUI:Create("Dropdown")
  difficultyDropdown:SetList(difficultyNames)
  
  difficultyDropdown:SetLabel("Difficulty")
  difficultyDropdown:SetRelativeWidth(0.2)
  difficultyDropdown:SetCallback("OnValueChanged", FilterDifficulty)
  
  encounterDropdown = AceGUI:Create("Dropdown")
  encounterDropdown:SetLabel("Encounter")
  encounterDropdown:SetRelativeWidth(0.4)
  encounterDropdown:SetCallback("OnValueChanged", FilterEncounter)
    
  instanceDropdown = AceGUI:Create("Dropdown")
  UpdateInstanceList()
  instanceDropdown:SetLabel("Instance")
  instanceDropdown:SetRelativeWidth(0.39)
  instanceDropdown:SetCallback("OnValueChanged", FilterInstance)
  
  difficultyDropdown:SetPoint("RIGHT", instanceDropdown.frame, 0, 0)
  encounterDropdown:SetPoint("RIGHT", difficultyDropdown.frame, 0, 0)
  
  container:AddChild(instanceDropdown)
  container:AddChild(difficultyDropdown)
  container:AddChild(encounterDropdown)
  
 
  local scrollContainer = AceGUI:Create("SimpleGroup")
  scrollContainer:SetFullWidth(true)
  scrollContainer:SetFullHeight(true)
  scrollContainer:SetLayout("Fill")
  container:AddChild(scrollContainer)
  
  scrollFrame = AceGUI:Create("ScrollFrame")
  scrollFrame:SetLayout("Flow")
  scrollContainer:AddChild(scrollFrame)
  
  FilterInstance(encounterDropdown, nil, 0)
end

local function DrawUserTab(container)
  local scrollContainer = AceGUI:Create("SimpleGroup")
  container.scrollContainer = scrollContainer
  scrollContainer:SetFullWidth(true)
  scrollContainer:SetFullHeight(true)
  scrollContainer:SetLayout("Fill")
  container:AddChild(scrollContainer)

  local scrollFrame = AceGUI:Create("ScrollFrame")
  container.scrollFrame = scrollFrame
  scrollFrame:SetLayout("Flow")
  scrollContainer:AddChild(scrollFrame)
  
  local users = {}
  for instance, difficulties in pairs(BossList_entries) do    -- Instances
    for difficulty, encounters in pairs(difficulties) do      -- Difficulties
      for encounter, items in pairs(encounters) do            -- Encounters
        for item, desires in pairs(items) do                  -- Items
          for desire, characters in pairs(desires) do         -- Desires
            for character, lastUpdate in pairs(characters) do -- Characters
              if users[character] == nil or users[character] < lastUpdate then
                users[character] = lastUpdate
              end
            end
          end
        end
      end
    end
  end
  
  local userCount = 0
  local elements = {}
  for player, lastUpdate in pairsByKeys(users) do
    local playerLabel = AceGUI:Create("Label")
    playerLabel:SetRelativeWidth(0.33)
    playerLabel:SetText(player)
    playerLabel.label:SetJustifyH("LEFT")
    tinsert(elements, playerLabel)
    
    local lastSyncLabel = AceGUI:Create("Label")
    lastSyncLabel:SetRelativeWidth(0.33)
    lastSyncLabel:SetText(BossList.lastSync[player] or "N/A")
    lastSyncLabel.label:SetJustifyH("CENTER")
    tinsert(elements, lastSyncLabel)
    
    local lastUpdateLabel = AceGUI:Create("Label")
    lastUpdateLabel:SetRelativeWidth(0.33)
    lastUpdateLabel:SetText(TimeDifference(lastUpdate))
    lastUpdateLabel:SetPoint("RIGHT", playerLabel.frame, 0,0)
    lastUpdateLabel.label:SetJustifyH("RIGHT")
    tinsert(elements, lastUpdateLabel)
    
    userCount = userCount + 1
  end
  
  local heading = AceGUI:Create("Heading")
  heading:SetFullWidth(true)
  heading:SetText(userCount .. " user(s) found")
    
  scrollFrame:AddChildren(heading, unpack(elements))
end

local function SelectGroup(container, event, group)
  container:ReleaseChildren()
  if group == "tab1" or group == "tab2" then
    DrawBossesTab(container)
  elseif group == "tab3" then
    DrawUserTab(container)
  end
end

function BossListFrame.init()
  local frame = AceGUI:Create("Frame")
  --frame.frame:SetParent(BossListFrame)
  --table.insert(UISpecialFrames, "BossListFrame")
  frame:SetTitle("BossList")
  frame:SetWidth(650)
  frame:SetCallback("OnClose", function(widget) AceGUI:Release(widget); tab = nil; BossListFrame.visible = false end)
  frame:SetLayout("Fill")
  BossListFrame.OverviewFrame = frame
  
  tab = AceGUI:Create("TabGroup")
  tab:SetLayout("Flow")
  tab:SetTabs({{text="Needs by bosses", value="tab1"}, {text="Needs by items", value="tab2"}, {text="Users", value="tab3"}})
  tab:SetCallback("OnGroupSelected", SelectGroup)
  tab:SelectTab("tab1")
  frame:AddChild(tab)
end


function BossList.addEntry(instanceID, difficulty, encounterID, itemID, desirability, characterName, lastUpdate)
  if BossList.IsBlackListed(characterName) then
    return
  end
  
  if BossList_entries[instanceID] == nil then BossList_entries[instanceID] = {} end
  local instanceTable = BossList_entries[instanceID]
  if instanceTable[difficulty] == nil then instanceTable[difficulty] = {} end
  local difficultyTable = instanceTable[difficulty]
  if difficultyTable[encounterID] == nil then difficultyTable[encounterID] = {} end
  local encounterTable = difficultyTable[encounterID]
  if encounterTable[itemID] == nil then encounterTable[itemID] = {} end
  local itemTable = encounterTable[itemID]
  if itemTable[desirability] == nil then itemTable[desirability] = {} end
  local desireTable = itemTable[desirability]
  
  -- Check for old entry in all desires
  for desire, characters in pairs(itemTable) do
    if characters[characterName] ~= nil then
      characters[characterName] = nil
    end
  end
  
  lastUpdate = lastUpdate or time()
  
  if time() - lastUpdate < 7776000 then -- Newer than 3 months
    desireTable[characterName] = lastUpdate
    
    --if visible then
    --  BossListFrame.Show()
    --end
  end
end

BossListFrame.Show = function(this)
  BossListFrame.visible = true
  if tab == nil then
    this.init()
  else
    if tab.localstatus.selected == "tab1" then
      UpdateInstanceList()
      UpdateEncounterList(curInstance)
      DrawEntries(scrollFrame, curInstance, curDifficulty, curEncounter)
    elseif tab.localstatus.selected == "tab2" then
      UpdateInstanceList()
      UpdateEncounterList(curInstance)
      DrawItems(scrollFrame, curInstance, curDifficulty, curEncounter)
    end
  end
  collectgarbage()
end

BossListFrame.Hide = function(this)
  if BossListFrame.OverviewFrame ~= nil then
    BossListFrame.OverviewFrame:Hide(); tab = nil; BossListFrame.visible = false;
  end
end

hooksecurefunc("CloseSpecialWindows", function()
  BossList.frame:Hide()
end)